/*  Sjaak, a program for playing chess variants
 *  Copyright (C) 2011, 2014  Evert Glebbeek
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MOVE_H
#define MOVE_H

#include <stdbool.h>
#include <stdint.h>
#include "pieces.h"
#include "assert.h"

/* Define structure to hold a move.
 * Because of the specific requirements of variants or other games, this has to be a fairly flexible data
 * structure. Each move can consist of (in that order):
 *  N pickups (N = 0, 1, 2, 3)
 *  N drops   (N = 0, 1, 2, 3)
 *  N store/retrieve (N = 0, 1)
 * A "pickup" stores a square and a piece to be picked up (7+4+1 = 12 bits).
 * A "drop" stores a square and a piece to be dropped there (7+4+1 = 12 bits).
 * A "store" stores a piece in the holdings (1+4+1 = 6 bits), a "retrieve" does the opposite
 * The extra bits for each piece are used to encode the colour of the piece.
 *
 * The order in which the components should be resolved is:
 *  1. pickups (so all destination squares are empty)
 *  2. drops (which may happen on the square that was just cleared by the move)
 *  3. store/retrieve (could be any order really)
 * To encode a promotion, then, we don't encode the pawn move: we pickup the pawn, and drop the promoted piece
 * on its destination square.
 *
 * Move bits:
 *  1-2  Number of pieces picked up
 *  3-4  Number of pieces dropped
 *  5    Number of pieces taken from/stored in holdings
 *  6-17 First pickup/drop/store/retrieve
 * 18-29 Second pickup/drop/store/retrieve
 * 30-41 Third pickup/drop/store/retrieve
 * 42-53 Fourth pickup/drop/store/retrieve
 * 54-59 Fifth store/retrieve
 *  60   Unused
 *  61   Unused
 *  62   Reset half-move counter (50 move counter)
 *  63   Don't change the side-to-move
 *  64   Set the en-passant target square
 *
 * TODO: We can fake a fifth "drop" by taking the square from the first
 * pickup and the piece from the holdings. This would allow us to store
 * three piece drops and two pickups alongside a retrieve, which is what is
 * needed to encode a castling-gating move in Seirawan chess.
 * A proper implementation would leave makemove unchanged and only affect
 * the storing of gating moves.
 *
 * TODO: change the move encoding. The present symmetry between
 * drops/pickups is nice, but not essential. If we rely on the board to
 * determine pickup pieces, we can instead use the following encoding:
 *  Pickup: square (7 bits)
 *  Drop:   piece/side/square (4+1+7=12 bits)
 *  Swap:   square/square (7+7=14 bits)
 * Normal moves then consist of a swap
 * Captures consist of a pickup and a swap
 * Promotions consist of one (or two, for captures) pickup(s) and a drop
 * Castling consists of two swaps
 * Gating moves add an extra drop.
 * A castle+gate in this case takes up 2*14+12 = 40 bits.
 * A capture promotion takes 2*7+12 = 26 bits
 * A regular capture takes 14+7 = 21 bits
 * To perform a move:
 *  1. Resolve pickups, store pieces in unmake info
 *  2. Perform swaps.
 *  3. Perform drops.
 * For printing of SAN moves it is convenient to have the "primary" piece
 * type as well.
 *  bits
 *   1-2  Number of pickups
 *   3-4  Number of "swaps"
 *   5-6  Number of drops
 *   7    Number of pieces taken/from holdings
 *   8-16 Unused/flags
 *  17-20 Primary piece
 *  21-64 (44 bits) move actions
 * We even have room to expand the number of piece types to 32, which makes
 * implementing variants with many pieces easier (or possible at all).
 */
typedef uint64_t move_t;

/* Offsets and test masks */
#define MOVE_SQUARE_MASK      0x007f
#define MOVE_PIECE_MASK       0x000f
#define MOVE_DROP_MASK        0x0fff
#define MOVE_PICKUP_MASK      0x0fff
#define MOVE_HOLDING_MASK     0x007f
#define MOVE_SWAP_MASK        0x3fff

#define MOVE_SET_ENPASSANT    0x8000000000000000ll
#define MOVE_KEEP_TURN        0x4000000000000000ll
#define MOVE_RESET50          0x2000000000000000ll

#define MOVE_HEADER_SIZE      7
#define MOVE_PICKUP_SIZE      12                // bits
#define MOVE_DROP_SIZE        MOVE_PICKUP_SIZE  // bits
#define MOVE_SWAP_SIZE        14

#define MOVE_SLOT1            (MOVE_HEADER_SIZE)
#define MOVE_SLOT2            (MOVE_SLOT1 + MOVE_PICKUP_SIZE)
#define MOVE_SLOT3            (MOVE_SLOT2 + MOVE_PICKUP_SIZE)
#define MOVE_SLOT4            (MOVE_SLOT3 + MOVE_PICKUP_SIZE)

extern char piece_symbol_string[MAX_PIECE_TYPES+1];
extern char piece_drop_string[MAX_PIECE_TYPES+1];

#ifdef __cplusplus
const char *move_to_string(move_t move, char *buffer = NULL);
const char *move_to_lan_string(move_t move, bool castle_san = true, char *buffer = NULL);
#endif

/**********************************************************************
 *                      static inline functions                       *
 **********************************************************************/
static inline int get_move_pickups(move_t move)
{
   return move & 0x03;
}

static inline int get_move_drops(move_t move)
{
   return (move >> 2) & 0x03;
}

static inline int get_move_swaps(move_t move)
{
   return (move >> 5) & 0x03;
}

static inline int get_move_holdings(move_t move)
{
   return (move >> 4) & 0x01;
}



static inline int encode_number_pickups(int count)
{
   assert(count <= 3);
   return count;
}

static inline int encode_number_drops(int count)
{
   assert(count <= 3);
   return count << 2;
}

static inline int encode_number_swaps(int count)
{
   assert(count <= 3);
   return count << 5;
}

static inline int encode_number_holdings(int count)
{
   assert(count <= 1);
   return count << 4;
}

static inline int encode_move_piece(int piece)
{
   return 0;
}

/* Get the first, second or third pickup */
static inline int get_move_pickup(move_t move, int n)
{
   assert (n<=get_move_pickups(move));
   int shift = n * MOVE_PICKUP_SIZE;

   return (move >> (MOVE_SLOT1 + shift)) & MOVE_PICKUP_MASK;
}

/* Get the first, second or third drop */
static inline int get_move_drop(move_t move, int n)
{
   assert (n<=get_move_drops(move));
   int shift = MOVE_DROP_SIZE*n + MOVE_PICKUP_SIZE*get_move_pickups(move);

   return (move >> (MOVE_SLOT1 + shift)) & MOVE_DROP_MASK;
}

static inline int get_move_swap(move_t move, int n)
{
   assert (n<=get_move_swaps(move));
   int shift = MOVE_SWAP_SIZE*n + MOVE_DROP_SIZE*get_move_drops(move) + MOVE_PICKUP_SIZE*get_move_pickups(move);

   return (move >> (MOVE_SLOT1 + shift)) & MOVE_SWAP_MASK;
}

static inline int get_move_holding(move_t move)
{
   return (move >> (MOVE_SLOT1 + MOVE_PICKUP_SIZE*(get_move_drops(move)+get_move_pickups(move)))) & MOVE_HOLDING_MASK;
}

static inline int decode_swap_from(int swap)
{
   return swap & MOVE_SQUARE_MASK;
}

static inline int decode_swap_to(int swap)
{
   return (swap >> 7) & MOVE_SQUARE_MASK;
}

/* Decode drop and pickup piece/squares */
static inline int decode_drop_square(int drop)
{
   return (drop >> 5) & MOVE_SQUARE_MASK;
}

static inline int decode_drop_piece(int drop)
{
   return drop & MOVE_PIECE_MASK;
}

static inline side_t decode_drop_side(int drop)
{
   return (side_t)((drop>>4) & 0x1);
}

/* The code for pickups is the same as for drops */
static inline int decode_pickup_square(int pickup)
{
   return (pickup >> 5) & MOVE_SQUARE_MASK;
}
#define decode_pickup_piece decode_drop_piece
#define decode_pickup_side decode_drop_side

/* The code for holdings is almost the same as for drops */
static inline int decode_holding_count(int hold)
{
   return decode_drop_square(hold) - 2;
}
#define decode_holding_piece decode_drop_piece
#define decode_holding_side decode_drop_side

static inline int encode_swap(int from, int to)
{
   return from | to << 7;
}

static inline int encode_pickup(int piece, int from)
{
   return piece | from<<5;
}

static inline int encode_drop(int piece, int to)
{
   return piece | to<<5;
}

/* Encode a normal (non-capture) move */
static inline move_t encode_normal_move(uint8_t piece, uint8_t from, uint8_t to)
{
   return encode_number_pickups(1) |
          encode_number_drops(1)   |
          encode_move_piece(piece) |
          encode_pickup(piece, from) << MOVE_SLOT1 |
          encode_drop(piece, to) << MOVE_SLOT2;
}

/* Encode an en-passant capture, where the pickup square is different from the destination square */
static inline move_t encode_en_passant_capture(uint8_t piece, uint8_t from, uint8_t to, uint8_t ptaken, uint8_t ep)
{
   return encode_number_pickups(2) |
          encode_number_drops(1) |
          encode_move_piece(piece) |
          encode_pickup(piece,  from) << MOVE_SLOT1 |
          encode_pickup(ptaken, ep)   << MOVE_SLOT2 | (uint64_t)
          encode_drop  (piece,  to)   << MOVE_SLOT3 | MOVE_RESET50;
}

/* Encode a normal capture */
static inline move_t encode_normal_capture(uint8_t piece, uint8_t from, uint8_t to, uint8_t ptaken)
{
   return encode_en_passant_capture(piece, from, to, ptaken, to);
}

/* Encode a normal (non-capture) promotion */
static inline move_t encode_normal_promotion(uint8_t piece, uint8_t from, uint8_t to, uint8_t tpiece)
{
   return encode_number_pickups(1) |
          encode_number_drops(1)   |
          encode_move_piece(piece) |
          encode_pickup(piece,  from) << MOVE_SLOT1 |
          encode_drop  (tpiece, to)   << MOVE_SLOT2 | MOVE_RESET50;
}

/* Encode a capture promotion */
static inline move_t encode_capture_promotion(uint8_t piece, uint8_t from, uint8_t to, uint8_t ptaken, uint8_t tpiece)
{
   return encode_number_pickups(2) |
          encode_number_drops(1) |
          encode_move_piece(piece) |
          encode_pickup(piece,  from) << MOVE_SLOT1 |
          encode_pickup(ptaken, to)   << MOVE_SLOT2 | (uint64_t)
          encode_drop  (tpiece, to)   << MOVE_SLOT3 | MOVE_RESET50;
}

/* Encode a castling move */
static inline move_t encode_castle_move(uint8_t piece, uint8_t from, uint8_t to, uint8_t p2, uint8_t f2, uint8_t t2) 
{
   return encode_number_pickups(2) |
          encode_number_drops(2) |
          encode_move_piece(piece) |
            (piece | from<<5) << MOVE_SLOT1 |
            (p2    | f2  <<5) << MOVE_SLOT2 | (uint64_t)
            (piece | to  <<5) << MOVE_SLOT3 | (uint64_t)
            (p2    | t2  <<5) << MOVE_SLOT4 | MOVE_RESET50;
}

/* Encode a drop */
static inline move_t encode_drop_move(uint8_t piece, uint8_t to)
{
   return encode_number_drops(1) | encode_move_piece(piece) | encode_drop(piece, to) << MOVE_SLOT1 | MOVE_RESET50;
}

/* Add store/retrieval information to a move */
static inline move_t add_move_store(move_t move, uint8_t piece, int8_t count)
{
   int shift;
   assert(count < 2);
   assert(count > -2);
   assert((get_move_pickups(move) + get_move_drops(move)) < 5);

   count += 2;
   shift = MOVE_SLOT1 +
           MOVE_DROP_SIZE * get_move_drops(move) +
           MOVE_SWAP_SIZE * get_move_swaps(move) +
           MOVE_PICKUP_SIZE * get_move_pickups(move);

   return move | encode_number_holdings(1) | (uint64_t)(piece | count<<5) << shift;
}
#define add_move_retrieve(move, piece, count) add_move_store((move), (piece), -(count))

/* Query move types:
 * A normal move is one pickup, one drop (by the same piece)
 * A capture is two pickups (from two different sides), one drop
 * A promotion is a drop by a different piece type than a capture
 * A drop has no pickups
 * A "shot" is a pickup without a drop
 * Castling is two pickups and two drops
 */

static inline bool is_capture_move(move_t move)
{
   return get_move_pickups(move) > get_move_drops(move);
}

static inline bool is_promotion_move(move_t move)
{
   if (get_move_drops(move) > 1 || get_move_pickups(move) < 1)
      return false;
   int drop_piece = decode_drop_piece(get_move_drop(move, 0));
   return drop_piece != decode_pickup_piece(get_move_pickup(move, 0));
}

static inline bool is_irreversible_move(move_t move)
{
   return (move & MOVE_RESET50) || is_capture_move(move) || is_promotion_move(move);
}

static inline bool is_drop_move(move_t move)
{
   return (get_move_pickups(move) == 0) && (get_move_drops(move) == 1);
}

static inline bool is_castle_move(move_t move)
{
   return (get_move_pickups(move) == 2) && (get_move_drops(move) == 2);
}

/* Convenience macros: the piece type of the first pickup and the first pickup and drop squares
 * By lucky coincidence, this works correctly for pure drop-moves too.
 */
static inline int get_move_piece(move_t move)
{
   uint16_t p = get_move_pickup(move, 0);

   return decode_pickup_piece(p);
}

static inline int get_move_captured_piece(move_t move)
{
   uint16_t p = get_move_pickup(move, 1);

   return decode_pickup_piece(p);
}

static inline int get_move_promotion_piece(move_t move)
{
   uint16_t p = get_move_drop(move, 0);

   return decode_drop_piece(p);
}

static inline int get_move_from(move_t move)
{
   uint16_t p = get_move_pickup(move, 0);

   return decode_pickup_square(p);
}

static inline int get_move_to(move_t move)
{
   uint16_t p = get_move_drop(move, 0);

   return decode_drop_square(p);
}

static inline int get_castle_move_from2(move_t move)
{
   uint16_t p = get_move_pickup(move, 1);

   return decode_pickup_square(p);
}

static inline int get_castle_move_to2(move_t move)
{
   uint16_t p = get_move_drop(move, 1);

   return decode_drop_square(p);
}

static inline int get_move_capture_square(move_t move)
{
   uint16_t p = get_move_pickup(move, 1);

   return decode_drop_square(p);
}

static inline int get_move_player(move_t move)
{
   uint16_t p = get_move_pickup(move, 0);

   return decode_pickup_side(p);
}

#endif

